package javachess;
import javachess.AISearchAgent;
import javachess.Board;
import javachess.OpeningBook;

/***************************************************************************
 * A sophisticated search agent
 *
 * Purpose: A (mostly) state-of-the-art search agent, implementing advanced
 * techniques like the iterative-deepening MTDF search algorithm, transposition
 * table, opening book and history table.
 *
 * History:
 * 05.10.00 Completed initial version
 **************************************************************************/

public class AISearchAgentMTDF extends AISearchAgent
{
  // A reference to the game's opening book
  private OpeningBook Openings;

  // A measure of the effort we are willing to expend on search
  private static final int MaxSearchSize = 50000;

  // Construction
  public AISearchAgentMTDF( OpeningBook ref )
  {
    super();
    Openings = ref;
  }

  /****************************************************************************
   * PUBLIC METHODS
   ***************************************************************************/
public static int countforget;
  // Move selection: An iterative-deepening paradigm calling MTD(f) repeatedly
  public Move PickBestMove( Board theBoard )
  {
    // First things first: look in the Opening Book, and if it contains a
    // move for this position, don't search anything
    MoveCounter++;
    Move Mov = null;
    Mov = Openings.Query( theBoard );
    if ( Mov != null )
      return Mov;

    // Store the identity of the moving side, so that we can tell Evaluator
    // from whose perspective we need to evaluate positions
    FromWhosePerspective = theBoard.GetCurrentPlayer();

    // Should we erase the history table?
 //   if ( ( Rnd.nextInt() % 6 ) == 2 )
    if ((++countforget % 6) == 2)
      HistTable.Forget();

    // Begin search.  The search's maximum depth is determined on the fly,
    // according to how much effort has been spent; if it's possible to search
    // to depth 8 in 5 seconds, then by all means, do it!
    int bestGuess = 0;
    int iterdepth = 1;
    long StartTime = 0;
    long Duration = 0;

    while( true )
    {
      // Searching to depth 1 is not very effective, so we begin at 2
      iterdepth++;

      // Compute efficiency statistics
      NumRegularNodes = 0; NumQuiescenceNodes = 0;
      NumRegularTTHits = 0; NumQuiescenceTTHits = 0;
      NumRegularCutoffs = 0; NumQuiescenceCutoffs = 0;

	  StartTime = System.currentTimeMillis();
      // Look for a move at the current depth
      Debug.println("MTDF (theBoard, " + bestGuess + ", " + iterdepth + (");"));
      
      // if (iterdepth < ) Debug.EnableOut = true; else Debug.EnableOut = false; 
      Mov = MTDF( theBoard, bestGuess, iterdepth );
      bestGuess = Mov.MoveEvaluation;
	  Debug.println("bestGuess = " + Mov.MoveEvaluation);
      Duration = System.currentTimeMillis() - StartTime;

      // Feedback!
	  System.out.print( "Iteration of depth " + iterdepth + "; best move = " );
	  Mov.Print();
	  System.out.print( "  --> Transposition Table hits for regular nodes: " );
	  System.out.println( NumRegularTTHits + " of " + NumRegularNodes );
	  System.out.print( "  --> Transposition Table hits for quiescence nodes: " );
	  System.out.println( NumQuiescenceTTHits + " of " + NumQuiescenceNodes );
	  System.out.println( "  --> Number of cutoffs for regular nodes: " + NumRegularCutoffs );
	  System.out.println( "  --> Number of cutoffs in quiescence search: " + NumQuiescenceCutoffs );
	  System.out.println( "  ---> nodes: " + (NumQuiescenceNodes + NumRegularNodes) + "  hits: " + (NumQuiescenceTTHits + NumRegularTTHits) + "  duration: " + Duration + "ms");
	  System.out.print( "  ---> nodes / seconds: ");
      if (Duration < 5) System.out.println("na"); 
        else System.out.println(((NumQuiescenceNodes - NumQuiescenceTTHits + NumRegularNodes - NumRegularTTHits) / Duration) + " kNodes !");

		Debug.print( "Iteration of depth " + iterdepth + "; best move = " );
		
		Debug.print( "Move: " );
		if ( Mov.MoveType == Move.MOVE_STALEMATE )
		{
		  Debug.println( "STALEMATE!!!" );
		}
		if ( Mov.MoveType != Move.MOVE_RESIGN )
		{
		  Debug.print( Board.PieceStrings[ Mov.MovingPiece ] );
		  Debug.print( "  [ " );
		  Debug.print( "" + Mov.SourceSquare );
		  Debug.print( ", " );
		  Debug.print( "" + Mov.DestinationSquare );
		  Debug.print( " ] TYPE: " );
		  Debug.println( "" + Mov.MoveType );
		}
		else
		{
		  Debug.println( "RESIGNATION!" );
		}
	  	
		
		Debug.print( "  --> Transposition Table hits for regular nodes: " );
		Debug.println( NumRegularTTHits + " of " + NumRegularNodes );
		Debug.print( "  --> Transposition Table hits for quiescence nodes: " );
		Debug.println( NumQuiescenceTTHits + " of " + NumQuiescenceNodes );
		Debug.println( "  --> Number of cutoffs for regular nodes: " + NumRegularCutoffs );
		Debug.println( "  --> Number of cutoffs in quiescence search: " + NumQuiescenceCutoffs );
		Debug.println( "  ---> nodes: " + (NumQuiescenceNodes + NumRegularNodes) + "  hits: " + (NumQuiescenceTTHits + NumRegularTTHits) + "  duration: " + Duration + "ms");
		Debug.print( "  ---> nodes / seconds: ");
		if (Duration < 5) Debug.println("na"); 
		  else Debug.println(((NumQuiescenceNodes - NumQuiescenceTTHits + NumRegularNodes - NumRegularTTHits) / Duration) + " kNodes !");
		Debug.println( "");
		Debug.println( "");

      // Get out if we have searched deep enough
      if ( ( NumRegularNodes + NumQuiescenceNodes ) > MaxSearchSize )
        break;
      if ( iterdepth >= 15 )
        break;
    }

    return Mov;
  }

  /***************************************************************************
   * PRIVATE METHODS
   **************************************************************************/

  // private Move MTDF
  // Use the MTDF algorithm to find a good move.  MTDF repeatedly calls
  // alphabeta with a zero-width search window, which creates very many quick
  // cutoffs.  If alphabeta fails low, the next call will place the search
  // window lower; in a sense, MTDF is a sort of binary search mechanism into
  // the minimax space.
  private Move MTDF( Board theBoard, int target, int depth )
  {
    int beta;
    Move Mov;
    int currentEstimate = target;
    int upperbound = ALPHABETA_MAXVAL;
    int lowerbound = ALPHABETA_MINVAL;

    // This is the trick: make repeated calls to alphabeta, zeroing in on the
    // actual minimax value of theBoard by narrowing the bounds
    do {
      Debug.println("  do");
      if ( currentEstimate == lowerbound ) {
        beta = currentEstimate + 1;
		Debug.println("    inc(beta); newbeta = " + beta);
      }
      else
        beta = currentEstimate;

	  Debug.println("    depth = " + depth + "  beta = " + beta);
	  Debug.println("    Mov = UnrolledAlphabeta( theBoard, depth, beta - 1, beta");

      Mov = UnrolledAlphabeta( theBoard, depth, beta - 1, beta );
      currentEstimate = Mov.MoveEvaluation;
	  Debug.println("    currentEstimate = " + Mov.MoveEvaluation + "  (beta = " + beta + ")");
		
      if ( currentEstimate < beta )
        upperbound = currentEstimate;
      else
        lowerbound = currentEstimate;
      Debug.println("    upperbound = " + upperbound + "  lowerbound = " + lowerbound); 
	  Debug.println("  while ( lowerbound < upperbound )");
    } while ( lowerbound < upperbound );
    Debug.println("  exit while");

    return Mov;
  }

  // private Move UnrolledAlphabeta
  // The standard alphabeta, with the top level "unrolled" so that it can
  // return a Move structure instead of a mere minimax value
  // See AISearchAgent.Alphabeta for detailed comments on this code
  private Move UnrolledAlphabeta( Board theBoard, int depth, int alpha,
                                    int beta )
  {
    Move BestMov = new Move();

    MoveListGenerator movegen = new MoveListGenerator();
    movegen.ComputeLegalMoves( theBoard );
  	HistTable.SortMoveList( movegen, theBoard.GetCurrentPlayer() );

    Board newBoard = new Board();
    int bestSoFar;

    bestSoFar = ALPHABETA_MINVAL;
    int currentAlpha = alpha;
    Move mov;

	Debug.println("      while = movegen.Next()");
    movegen.TracePrint();
    // Loop on the successors
    while( ( mov = movegen.Next() ) != null )
    {
      Debug.println("      --nextmove: [" + mov.SourceSquare + ", " + mov.DestinationSquare + "]");
      // Compute a board position resulting from the current successor
      newBoard.Clone( theBoard );
      newBoard.ApplyMove( mov );

      // And search it in turn
	  Debug.println("        MINNODE = " +MINNODE+ " (<-MINNODE)  newBoard, depth = " +depth+ "  currentAlpha = " +currentAlpha+ "  = beta = " + beta);
	  Debug.println("        AlphaBeta( MINNODE, newBoard, depth - 1, currentAlpha, beta)");
      int movScore = AlphaBeta( MINNODE, newBoard, depth - 1, currentAlpha, beta );
	  Debug.println("        movescore = " + movScore);

      // Ignore illegal moves in the alphabeta evaluation
      if ( movScore == ALPHABETA_ILLEGAL ) {
		Debug.println("        illegal movescore");
        continue;
      }
      currentAlpha = Math.max( currentAlpha, movScore );
	  Debug.println("        currentAlpha = max( currentAlpha = " + currentAlpha + ", movScore = " + movScore);

		
      // Is the current successor better than the previous best?
      if ( movScore > bestSoFar )
      {
		Debug.println("        movScore ("+movScore+") > bestSoFar ("+bestSoFar+")");
		Debug.println("        NEWBESTMOVE");
        BestMov.Copy( mov );
        bestSoFar = movScore;
        BestMov.MoveEvaluation = bestSoFar;

        // Can we cutoff now?
        if ( bestSoFar >= beta )
        {
  		  Debug.println("        bestSoFar ("+bestSoFar+") >= beta ("+beta+") --> CUTOFF");
		  Debug.println("          TransTable.StoreBoard + HistoryTable.AddCount");
          TransTable.StoreBoard( theBoard, bestSoFar, Move.EVALTYPE_UPPERBOUND, depth, MoveCounter );

          // Add this move's efficiency in the historyTable
          HistTable.AddCount( theBoard.GetCurrentPlayer(), mov );
          return BestMov;
        }
      }
    }

    // Test for checkmate or stalemate
    if ( bestSoFar <= ALPHABETA_GIVEUP )
    {
      newBoard.Clone( theBoard );
      MoveListGenerator secondary = new MoveListGenerator();
      newBoard.SwitchSides();
      if ( secondary.ComputeLegalMoves( newBoard ) )
      {
        // Then, we are not in check and may continue our efforts.
        HistTable.SortMoveList( movegen, newBoard.GetCurrentPlayer() );
        movegen.ResetIterator();
        BestMov.MoveType = Move.MOVE_STALEMATE;
        BestMov.MovingPiece = Board.KING + theBoard.GetCurrentPlayer();
        while( ( mov = movegen.Next() ) != null )
        {
          newBoard.Clone( theBoard );
          newBoard.ApplyMove( mov );
          if ( secondary.ComputeLegalMoves( newBoard ) )
          {
            BestMov.MoveType = Move.MOVE_RESIGN;
          }
        }
      }
      else
      {
        // We're in check and our best hope is GIVEUP or worse, so either we are
        // already checkmated or will be soon, without hope of escape
        BestMov.MovingPiece = Board.KING + theBoard.GetCurrentPlayer();
        BestMov.MoveType = Move.MOVE_RESIGN;
      }
    }

    // If we haven't returned yet, we have found an accurate minimax score
    // for a position which is neither a checkmate nor a stalemate
    TransTable.StoreBoard( theBoard, bestSoFar, Move.EVALTYPE_ACCURATE, depth, MoveCounter );

    return BestMov;
  }
  
  public static boolean CoordToBit(String CoordStr, int SourceField, int DestField) {
	if (CoordStr.length() < 4) return false;
  	
	if (CoordStr.toLowerCase().substring(0, 4) == "resi") {
		SourceField = -1;
		DestField = -1;
		return true;
	}
  	
	int row, column;
  	
	column = (int)CoordStr.toLowerCase().charAt(0) - (int)'a';
	row    = (int)CoordStr.toLowerCase().charAt(1) - (int)'0';
	if (column < 0 || column > 7 || row < 0 || row > 7) {
		return false;   
	}
	SourceField = column + row * 8;
	

	column = (int)CoordStr.toLowerCase().charAt(0) - (int)'a';
	row    = (int)CoordStr.toLowerCase().charAt(1) - (int)'0';
	if (column < 0 || column > 7 || row < 0 || row > 7) {
		return false;   
	}
	DestField = column + row * 8;
  	
	return false;
  }
  
  public static boolean CoordToBit(String CoordStr, String BitString) {
	if (CoordStr.length() < 4) return false;
	
	return false;
  }
}