/*************************************************************************
 * HistoryTable - A heuristic used to pick an order of evaluation for moves
 *
 * The history heuristic is an extension of the old "killer move" system: if
 * a move has caused a lot of cutoffs recently, it will be tried early in the
 * hope that it will do so again.
 *
 * Using the history table is a gamble.  We could do without it entirely,
 * compute "successor positions" for each possible moves, look them up in
 * the transposition table and hope to get a cutoff this way, which would
 * insure fast cutoffs whenever possible.  On the other hand, HistoryTable
 * has no knowledge of the contents of the transposition table, so it may
 * cause a deep search of several moves even though another one would result
 * in an immediate cutoff...  However, History requires far less memory
 * and computation than creating a ton of successor Board objects, so we
 * hope that, on average, it will still be more efficient overall.
 *
 * History
 * 14.08.00 Creation
 ************************************************************************/

package javachess;
import javachess.MoveListGenerator;
import javachess.Move;
import java.util.*;

public class HistoryTable
{
  /***********************************************************************
   * DATA MEMBERS
   **********************************************************************/

  // the table itself; a separate set of cutoff counters exists for each
  // side
  int History[][][];
  int CurrentHistory[][];

  // This is a singleton class; the same history can be shared by two AI's
  private static HistoryTable theInstance;

  // A comparator, used to sort the moves
  private jcMoveComparator MoveComparator;

  /***********************************************************************
   * STATIC BLOCK
   ***********************************************************************/
  static
  {
    theInstance = new HistoryTable();
  }

  /***********************************************************************
   *  jcMoveComparator - Inner class used in sorting moves
   **********************************************************************/
   class jcMoveComparator implements Comparator
   {
     public int compare( Object o1, Object o2 )
     {
       Move mov1 = (Move) o1;
       Move mov2 = (Move) o2;
       if ( CurrentHistory[ mov1.SourceSquare ][ mov1.DestinationSquare ] >
            CurrentHistory[ mov2.SourceSquare ][ mov2.DestinationSquare ] )
         return -1;
       else
         return 1;
     }
   }

  /************************************************************************
   * PUBLIC METHODS
   ***********************************************************************/

  // Accessor
  public static HistoryTable GetInstance()
  {
    return theInstance;
  }

  // Sort a list of moves, using the Java "Arrays" class as a helper
  public boolean SortMoveList( MoveListGenerator theList, int movingPlayer )
  {
    // Which history will we use?
    CurrentHistory = History[ movingPlayer ];

    // Arrays can't sort a dynamic array like MoveListGenerator's ArrayList
    // member, so we have to use an intermediate.  Annoying and not too clean,
    // but it works...
    
    //Arrays.sort( theList.GetMoveList().toArray(), 0, theList.Size(), MoveComparator );
    BubbleSortMoves( theList.GetMoveList() );
    return true;
  }

	int CompareMoves( Move mov1, Move mov2)
	{
		if ( CurrentHistory[ mov1.SourceSquare ][ mov1.DestinationSquare ] >=
    	   CurrentHistory[ mov2.SourceSquare ][ mov2.DestinationSquare ] )
			return -1;
		else
			return 1;
	}


	void SwapMoves( Move mov1,  Move mov2)
	{
		// Swap Moves
		Move temp = new Move();

		temp.Copy( mov2 );						
		mov2.Copy( mov1 ); 
		mov1.Copy( temp );						
	}


	boolean BubbleSortMoves( ArrayList list  )
	{
		boolean changed;
	
		if (list.size() >= 2)
			while ( true )
				{
				changed = false;
			
				Move mov1;
				Move mov2;

				for ( int i = 1; i < list.size(); i++ )
					{
					mov1 = (Move) list.get( i - 1 );
					mov2 = (Move) list.get( i );
					if ( CompareMoves( mov1, mov2 ) > 0 )
						{
						changed = true;
						SwapMoves( mov1 , mov2 );
						}
					}
					
				if ( changed == false ) break;
				}

	return true;
	}
	
	
  // History table compilation
  public boolean AddCount( int whichPlayer, Move mov )
  {
    History[ whichPlayer ][ mov.SourceSquare ][ mov.DestinationSquare ]++;
    return true;
  }


  // public boolean Forget
  // Once in a while, we must erase the history table to avoid ordering
  // moves according to the results of very old searches
  public boolean Forget()
  {
    for( int i = 0; i < 2; i++ )
      for( int j = 0; j < 64; j++ )
        for( int k = 0; k < 64; k++ )
          History[ i ][ j ][ k ] = 0;
    return true;
  }

  /************************************************************************
   * PRIVATE METHODS
   ***********************************************************************/
  private HistoryTable()
  {
    History = new int[ 2 ][ 64 ][ 64 ];
    MoveComparator = new jcMoveComparator();
  }
  
  public void Print() {
	System.out.println("HistoryTable:");
	for (int side = 0; side < 2; side++) {
		if (side == Player.SIDE_WHITE) {
			System.out.println("Table white:");
		} else {
			System.out.println("Table black:");
		}
		for (int y = 0; y < 64; y++)
			for (int x = 0; x < 64; x++)
				if (History[side][x][y] != 0) {
					System.out.println("[" + x + "][" +  y + " " + History[side][x][y]);
				}
	}
  }
}