// AISearchAgent.cpp: Implementierung der Klasse AISearchAgent.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "AISearchAgent.h"
#include "MoveListGenerator.h"
#include "AISearchAgentAlphabeta.h"
#include "AISearchAgentMTDF.h"

/***************************************************************************
 * Static Class members
 **************************************************************************/
#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#endif

//protected:
  /** 
   * Search node types: MAXNODEs are nodes where the computer player is the
   * one to move
   */
  const bool AISearchAgent::MAXNODE = true;
  /** 
   * Search node types: MINNODEs are positions where the opponent is to move.
   */
  const bool AISearchAgent::MINNODE = false;

  // Alphabeta search boundaries
  const int AISearchAgent::ALPHABETA_MAXVAL  =  30000;
  const int AISearchAgent::ALPHABETA_MINVAL  = -30000;
  const int AISearchAgent::ALPHABETA_ILLEGAL = -31000;

  // Alphabeta search cutoff types
  const int AISearchAgent::ALPHABETA_NOCUT      =  0;
  const int AISearchAgent::ALPHABETA_BETACUT    =  1;
  const int AISearchAgent::ALPHABETA_BELOWALPHA =  2;
  const int AISearchAgent::ALPHABETA_LOWCUT     =  3;
  const int AISearchAgent::ALPHABETA_HIGHCUT    =  4;


  /** An approximate upper bound on the total value of all positional
   * terms in the evaluation function
   */
  const int AISearchAgent::EVAL_THRESHOLD = 200;

  /**
   * A score below which we give up: if Alphabeta ever returns a value lower
   * than this threshold, then all is lost and we might as well resign.  Here,
   * the value is equivalent to "mated by the opponent in 3 moves or less".
   */
  const int AISearchAgent::ALPHABETA_GIVEUP = -29995;

//public:
  // ID's for concrete subclasses; AISearchAgent works as a factory for its
  // concrete subclasses
  const int AISearchAgent::AISEARCH_ALPHABETA = 0;
  const int AISearchAgent::AISEARCH_MTDF	  = 1;
  const int AISearchAgent::AISEARCH_NEGASCOUT = 2;

/***************************************************************************
 * Constructor / Destructor
 **************************************************************************/

AISearchAgent::AISearchAgent()
{
    Evaluator = new BoardEvaluator();
    MoveCounter			 = 0;
	NumRegularNodes		 = 0;
    NumQuiescenceNodes   = 0;
    NumRegularTTHits     = 0;
    NumQuiescenceTTHits  = 0;
    NumRegularCutoffs    = 0;
    NumQuiescenceCutoffs = 0;
	FromWhosePerspective = 0;
	TransTables = ( TranspositionTable** ) malloc( 2 * sizeof ( TranspositionTable* ) );
	TransTables[ 0 ] = &TransTable;
	TransTables[ 1 ] = &TransTableBlack;
	TransTable.SetSide( Player::SIDE_WHITE );
	TransTableBlack.SetSide( Player::SIDE_BLACK );
}


AISearchAgent::~AISearchAgent()
{
	delete Evaluator;
}


/***************************************************************************
 * Operations
 **************************************************************************/

/**
 * Pick a function which the agent will use to assess the potency of a
 * position.  This may change during the game; for example, a special
 * "mop-up" evaluator may replace the standard when it comes time to drive
 * a decisive advantage home at the end of the game.
 */
bool AISearchAgent::AttachEvaluator( BoardEvaluator* eval )
{
	Evaluator = eval;
	return true;
}

/**
 * The basic alpha-beta algorithm, used in one disguise or another by
 * every search agent class
 */
int AISearchAgent::AlphaBeta( bool nodeType, Board* theBoard, int depth,
	                          int alpha, int beta )
{
    Move mov;

	Move Bestmov;

    // Count the number of nodes visited in the full-width search
    NumRegularNodes++;

    // First things first: let's see if there is already something useful
    // in the transposition table, which might save us from having to search
    // anything at all
    if ( TransTables[ theBoard->CurrentPlayer ]->LookupBoard( theBoard, &mov ) && ( mov.SearchDepth >= depth ) ) {
		if ( nodeType == MAXNODE ) {
			if ( ( mov.MoveEvaluationType == Move::EVALTYPE_ACCURATE ) ||
			     ( mov.MoveEvaluationType == Move::EVALTYPE_LOWERBOUND ) ) {
				if ( mov.MoveEvaluation >= beta ) {
					NumRegularTTHits++;
					return mov.MoveEvaluation;
				}
			}
		} else  { // MinNode
			if ( ( mov.MoveEvaluationType == Move::EVALTYPE_ACCURATE ) ||
			     ( mov.MoveEvaluationType == Move::EVALTYPE_UPPERBOUND ) ) {
				if ( mov.MoveEvaluation <= alpha ) {
					NumRegularTTHits++;
					return mov.MoveEvaluation;
				}
			}
		}
    }

    // If we have reached the maximum depth of the search, stop recursion
    // and begin quiescence search
    if ( depth == 0 ) {
		return QuiescenceSearch( nodeType, theBoard, alpha, beta );
    }

    // Otherwise, generate successors and search them in turn
    // If ComputeLegalMoves returns false, then the current position is illegal
    // because one or more moves could capture a king!
    // In order to slant the computer's strategy in favor of quick mates, we
    // give a bonus to king captures which occur at shallow depths, i.e., the
    // more plies left, the better.  On the other hand, if you are losing, it
    // really doesn't matter how fast...
    MoveListGenerator movegen;
    if ( !movegen.ComputeLegalMoves( theBoard ) )
		return ALPHABETA_ILLEGAL;

    // Sort the moves according to History heuristic values
    CutoffHist.SortMoveList( &movegen, theBoard->GetCurrentPlayer() );


    // OK, now, get ready to search
    Board newBoard;
    int bestSoFar;

	Move* pmov = NULL;

    // Case #1: We are searching a Max Node
    if ( nodeType == MAXNODE ) {
		bestSoFar = ALPHABETA_MINVAL;
		int currentAlpha = alpha;

		// Loop on the successors
		while( ( pmov = movegen.Next() ) != NULL ) {
			// Compute a board position resulting from the current successor
			newBoard.Clone( theBoard );
			newBoard.ApplyMove( pmov );

			// And search it in turn
			int movScore = AlphaBeta( !nodeType, &newBoard, depth - 1, currentAlpha, beta );
			// Ignore illegal moves in the alphabeta evaluation
			if ( movScore == ALPHABETA_ILLEGAL ) 
					continue;

			currentAlpha = max( currentAlpha, movScore );

			// Is the current successor better than the previous best?
			if ( movScore > bestSoFar ) {
				Bestmov.Set( pmov );
				bestSoFar = movScore;

				// Can we cutoff now?
				if ( bestSoFar >= beta ) {
					// Store this best move in the TransTable
					TransTables[ theBoard->CurrentPlayer ]->StoreBoard( theBoard, bestSoFar, Move::EVALTYPE_LOWERBOUND, depth, MoveCounter, &Bestmov, NumRegularNodes );

					// Add this move's efficiency in the CutoffTable
					CutoffHist.AddCount( theBoard->GetCurrentPlayer(), &Bestmov );
					NumRegularCutoffs++;

					return bestSoFar;
				}
			}
		}

		// Test for checkmate or stalemate
		// Both cases occur if and only if there is no legal move for MAX, i.e.,
		// if "bestSoFar" is ALPHABETA_MINVAL.  There are two cases: we
		// have checkmate ( in which case the score is accurate ) or stalemate ( in
		// which case the position should be re-scored as a draw with value 0.
		if ( bestSoFar <= ALPHABETA_MINVAL ) {
			// Can MIN capture MAX's king?  First, ask the machine to generate
			// moves for MIN
			newBoard.Clone( theBoard );
			if( newBoard.GetCurrentPlayer() == FromWhosePerspective ) 
				newBoard.SwitchSides();

			// And if one of MIN's moves is a king capture, indicating that the
			// position is illegal, we have checkmate and must return MINVAL.  We
			// add the depth simply to "favor" delaying tactics: a mate in 5 will
			// score higher than a mate in 3, because the likelihood that the
			// opponent will miss it is higher; might as well make life difficult!
			if ( !movegen.ComputeLegalMoves( &newBoard ) )
				return bestSoFar + depth;
				else
				return 0;
		}
    } else { // Case #2: Min Node
		bestSoFar = ALPHABETA_MAXVAL;
		int currentBeta = beta;
		while( ( pmov = movegen.Next() ) != NULL ) {
			newBoard.Clone( theBoard );
			newBoard.ApplyMove( pmov );

			int movScore = AlphaBeta( !nodeType, &newBoard, depth - 1, alpha, currentBeta );
			if ( movScore == ALPHABETA_ILLEGAL )
				continue;
			currentBeta = min( currentBeta, movScore );
			if ( movScore < bestSoFar ) {
				Bestmov.Set( pmov );
				bestSoFar = movScore;

				// Cutoff?
				if ( bestSoFar <= alpha ) {
					TransTables[ theBoard->CurrentPlayer ]->StoreBoard( theBoard, bestSoFar, Move::EVALTYPE_UPPERBOUND, depth, MoveCounter, &Bestmov, NumRegularNodes);
					CutoffHist.AddCount( theBoard->GetCurrentPlayer(), pmov );
					NumRegularCutoffs++;
			
					return bestSoFar;
				}
			}
		}
		// Test for checkmate or stalemate
		if ( bestSoFar >= ALPHABETA_MAXVAL ) {
			// Can MAX capture MIN's king?
			newBoard.Clone( theBoard );
			if( newBoard.GetCurrentPlayer() != FromWhosePerspective )
				newBoard.SwitchSides();
			if ( !movegen.ComputeLegalMoves( &newBoard ) )
				return bestSoFar + depth;
				else
				return 0;
		}
    }

    // If we haven't returned yet, we have found an accurate minimax score
    // for a position which is neither a checkmate nor a stalemate
    TransTables[ theBoard->CurrentPlayer ]->StoreBoard( theBoard, bestSoFar, Move::EVALTYPE_ACCURATE, depth, MoveCounter, &Bestmov, NumRegularNodes );
    return bestSoFar;
}


/**
 * A slight variant of alphabeta which only considers captures and null moves
 * This is necesary because the evaluation function can only be applied to
 * "quiet" positions where the tactical situation ( i.e., material balance ) is
 * unlikely to change in the near future.
 * Note that, in this version of the code, the quiescence search is not limited
 * by depth; we continue digging for as long as we can find captures.  Some other
 * programs impose a depth limit for time-management purposes.
 */
int AISearchAgent::QuiescenceSearch( bool nodeType, Board* theBoard, int alpha, int beta )
{
    Move mov;
    NumQuiescenceNodes++;

    // First things first: let's see if there is already something useful
    // in the transposition table, which might save us from having to search
    // anything at all
    if ( TransTables[ theBoard->CurrentPlayer ]->LookupBoard( theBoard, &mov ) ) {
		if ( nodeType == MAXNODE ) {
			if ( ( mov.MoveEvaluationType == Move::EVALTYPE_ACCURATE ) ||
			     ( mov.MoveEvaluationType == Move::EVALTYPE_LOWERBOUND ) ) {
				if ( mov.MoveEvaluation >= beta ) {
					NumQuiescenceTTHits++;
		            return mov.MoveEvaluation;
				}
			}
		} else {
			if ( ( mov.MoveEvaluationType == Move::EVALTYPE_ACCURATE ) ||
			     ( mov.MoveEvaluationType == Move::EVALTYPE_UPPERBOUND ) ) {
				if ( mov.MoveEvaluation <= alpha ) {
					NumQuiescenceTTHits++;
					return mov.MoveEvaluation;
				}
			}
		}
    }

    int bestSoFar = ALPHABETA_MINVAL;

    // Start with evaluation of the null-move, just to see whether it is more
    // effective than any capture, in which case we must stop looking at
    // captures and damaging our position
	bestSoFar = Evaluator->EvaluateComplete( theBoard, FromWhosePerspective );
    if ( ( bestSoFar > ( beta + EVAL_THRESHOLD ) ) || ( bestSoFar < ( alpha - EVAL_THRESHOLD ) ) )
		return bestSoFar;

    // Now, look at captures
    MoveListGenerator movegen;
    if ( !movegen.ComputeQuiescenceMoves( theBoard ) )
		return bestSoFar;

    Board newBoard ; 
	Move *pmov;			//pointer to move object
	Move Bestmov;

    // Case #1: We are searching a Max Node
    if ( nodeType == AISearchAgent::MAXNODE ) {
		int currentAlpha = alpha;
		// Loop on the successors
		while( ( pmov = movegen.Next() ) != NULL ) {
			// Compute a board position resulting from the current successor
			newBoard.Clone( theBoard );
			newBoard.ApplyMove( pmov );

			// And search it in turn
			int movScore = QuiescenceSearch( !nodeType, &newBoard, currentAlpha, beta );
			// Ignore illegal moves in the alphabeta evaluation
			if ( movScore == ALPHABETA_ILLEGAL )
				continue;
			currentAlpha = max( currentAlpha, movScore );

			// Is the current successor better than the previous best?
			if ( movScore > bestSoFar ) {
				Bestmov.Set( pmov );
				bestSoFar = movScore;
				// Can we cutoff now?
				if ( bestSoFar >= beta ) {
					TransTables[ theBoard->CurrentPlayer ]->StoreBoard( theBoard, bestSoFar, Move::EVALTYPE_LOWERBOUND, 0, MoveCounter, &Bestmov, NumRegularNodes );
					// Add this move's efficiency in the HistoryTable
					CutoffHist.AddCount( theBoard->GetCurrentPlayer(), pmov );
					NumQuiescenceCutoffs++;
					return bestSoFar;
				}
			}
		}

		// Test for checkmate or stalemate
		// Both cases occur if and only if there is no legal move for MAX, i.e.,
		// if "bestSoFar" is ALPHABETA_MINVAL.  There are two cases: we
		// have checkmate ( in which case the score is accurate ) or stalemate ( in
		// which case the position should be re-scored as a draw with value 0.
		if ( bestSoFar <= ALPHABETA_MINVAL ) {
			// Can MIN capture MAX's king?  First, ask the machine to generate
			// moves for MIN
			newBoard.Clone( theBoard );
			if( newBoard.GetCurrentPlayer() == FromWhosePerspective )
			newBoard.SwitchSides();
			// And if one of MIN's moves is a king capture, indicating that the
			// position is illegal, we have checkmate and must return MINVAL.  We
			// add the depth simply to "favor" delaying tactics: a mate in 5 will
			// score higher than a mate in 3, because the likelihood that the
			// opponent will miss it is higher; might as well make life difficult!
			if ( !movegen.ComputeLegalMoves( &newBoard ) )
				return bestSoFar;
				else
				return 0;
		}
    } else { // Case #2: Min Node
		int currentBeta = beta;
		while( ( pmov = movegen.Next() ) != NULL ) {
			newBoard.Clone( theBoard );
			newBoard.ApplyMove( pmov );

			int movScore = QuiescenceSearch( !nodeType, &newBoard, alpha, currentBeta );
			if ( movScore == ALPHABETA_ILLEGAL )
				continue;
			currentBeta = min( currentBeta, movScore );
			if ( movScore < bestSoFar ) {
				Bestmov.Set( pmov );
				bestSoFar = movScore;
				// Cutoff?
				if ( bestSoFar <= alpha ) {
					TransTables[ theBoard->CurrentPlayer ]->StoreBoard( theBoard, bestSoFar, Move::EVALTYPE_UPPERBOUND, 0, MoveCounter, &Bestmov, NumRegularNodes );
					CutoffHist.AddCount( theBoard->GetCurrentPlayer(), pmov );
					NumQuiescenceCutoffs++;
					return bestSoFar;
				}
			}
		}
		// Test for checkmate or stalemate
		if ( bestSoFar >= ALPHABETA_MAXVAL ) {
			// Can MAX capture MIN's king?
			newBoard.Clone( theBoard );
			if( newBoard.GetCurrentPlayer() != FromWhosePerspective )
				newBoard.SwitchSides();
			if ( !movegen.ComputeLegalMoves( &newBoard ) )
				return bestSoFar;
				else
				return 0;
		}
	}

	// If we haven't returned yet, we have found an accurate minimax score
	// for a position which is neither a checkmate nor a stalemate
	TransTables[ theBoard->CurrentPlayer ]->StoreBoard( theBoard, bestSoFar, Move::EVALTYPE_ACCURATE, 0, MoveCounter, &Bestmov, NumRegularNodes );
	return bestSoFar;
}


/**
 * Standard "subclass factory" design pattern
 */
AISearchAgent* AISearchAgent::MakeNewAgent( int type, OpeningBook* ref )
{
    switch( type ) {
		case AISEARCH_ALPHABETA:
			return( new AISearchAgentAlphabeta() );
		case AISEARCH_MTDF:
			return( new AISearchAgentMTDF( ref ) );
		case AISEARCH_NEGASCOUT:
			return( new AISearchAgentNegaScout( ref ) );
		default:
			return NULL;
    }
}



void AISearchAgent::PickBestMove( Board* theBoard, Move* bestMove )
{
	return;
}
	

// binary read for AISearchAgent
istream& AISearchAgent::read( istream& is )
{
	is.read( ( char* )&MoveCounter, sizeof( MoveCounter ) );
	
	TransTable.Clear();
	TransTableBlack.Clear();
	CutoffHist.Forget();
	// enable the following lines are debugging options
	// CutoffHist.read( is );
	// TransTable.read( is );
	
	return is;
}

// binary write for AISearchAgent
ostream& AISearchAgent::write( ostream& os ) const
{
	os.write( ( char* )&MoveCounter, sizeof( MoveCounter ) );
	
	// enable the following lines are debugging options
	// CutoffHist.write( os );
	// TransTable.write( os );
	
	return os;
}




int AISearchAgent::MiniMax(bool nodeType, Board *theBoard, int depth)
{
	int g;
	Move* c;
	MoveListGenerator movegen;
	Board newBoard;

	NumRegularNodes++;

	if (depth == 0) return Evaluator->EvaluateComplete(theBoard, FromWhosePerspective);

	else if (nodeType == AISearchAgent::MAXNODE) {
		g = AISearchAgent::ALPHABETA_MINVAL;

		if ( !movegen.ComputeLegalMoves( theBoard ) )
			return AISearchAgent::ALPHABETA_ILLEGAL;

		while (c = movegen.Next()) {
			newBoard.Clone( theBoard );
			newBoard.ApplyMove( c );
			int newscore = MiniMax(!nodeType, &newBoard, depth-1);
			g = max(g, newscore);
		}
	} else { //minnode
		g = AISearchAgent::ALPHABETA_MAXVAL;

		if ( !movegen.ComputeLegalMoves( theBoard ) )
			return AISearchAgent::ALPHABETA_ILLEGAL;

		while (c = movegen.Next()) {
			newBoard.Clone( theBoard );
			newBoard.ApplyMove( c );
			int newscore = MiniMax(!nodeType, &newBoard, depth-1);
			g = min(g, newscore);
		}
	}

	return g;
}

int AISearchAgent::BasicAlphaBeta(bool nodeType, Board *theBoard, int depth, int alpha, int beta, bool sort)
{
	int g;
	Move* c;
	MoveListGenerator movegen;
	Board newBoard;

	NumRegularNodes++;

	if (depth == 0) return Evaluator->EvaluateComplete(theBoard, FromWhosePerspective);

	if ( !movegen.ComputeLegalMoves( theBoard ) )
		return AISearchAgent::ALPHABETA_ILLEGAL;
    // Sort the moves according to History heuristic values
    if (sort) CutoffHist.SortMoveList( &movegen, theBoard->GetCurrentPlayer() );

	if (nodeType == AISearchAgent::MAXNODE) {
		g = AISearchAgent::ALPHABETA_MINVAL;

		while ((g < beta) && (c = movegen.Next())) {
			newBoard.Clone( theBoard );
			newBoard.ApplyMove( c );
			int newscore = BasicAlphaBeta(!nodeType, &newBoard, depth-1, alpha, beta, sort);
			g = max(g, newscore);
			alpha = max(alpha, g);
			if (g >= beta) 
				CutoffHist.AddCount( theBoard->GetCurrentPlayer(), c );
		}
	} else { //minnode
		g = AISearchAgent::ALPHABETA_MAXVAL;

		while ((g > alpha) && (c = movegen.Next())) {
			newBoard.Clone( theBoard );
			newBoard.ApplyMove( c );
			int newscore = BasicAlphaBeta(!nodeType, &newBoard, depth-1, alpha, beta, sort);
			g = min(g, newscore);
			beta = min(beta, newscore);
			if (g <= alpha) 
				CutoffHist.AddCount( theBoard->GetCurrentPlayer(), c );
		}
	}

	return g;
}

int AISearchAgent::BasicAlphaBetaTT( bool nodeType, Board* theBoard, int depth,
	                          int alpha, int beta, bool sort )
{
    Move mov;

	Move Bestmov;

    // Count the number of nodes visited in the full-width search
    NumRegularNodes++;

    // First things first: let's see if there is already something useful
    // in the transposition table, which might save us from having to search
    // anything at all

    if ( TransTable.LookupBoard( theBoard, &mov ) && ( mov.SearchDepth >= depth ) ) {
		if ( nodeType == MAXNODE ) {
			if ( ( mov.MoveEvaluationType == Move::EVALTYPE_ACCURATE ) ||
			     ( mov.MoveEvaluationType == Move::EVALTYPE_LOWERBOUND ) ) {
				if ( mov.MoveEvaluation >= beta ) {
					NumRegularTTHits++;
					return mov.MoveEvaluation;
				}
			}
		} else  { // MinNode
			if ( ( mov.MoveEvaluationType == Move::EVALTYPE_ACCURATE ) ||
			     ( mov.MoveEvaluationType == Move::EVALTYPE_UPPERBOUND ) ) {
				if ( mov.MoveEvaluation <= alpha ) {
					NumRegularTTHits++;
					return mov.MoveEvaluation;
				}
			}
		}
    }

	if (depth == 0) return Evaluator->EvaluateComplete(theBoard, FromWhosePerspective);

    // Otherwise, generate successors and search them in turn
    // If ComputeLegalMoves returns false, then the current position is illegal
    // because one or more moves could capture a king!
    // In order to slant the computer's strategy in favor of quick mates, we
    // give a bonus to king captures which occur at shallow depths, i.e., the
    // more plies left, the better.  On the other hand, if you are losing, it
    // really doesn't matter how fast...
    MoveListGenerator movegen;
    if ( !movegen.ComputeLegalMoves( theBoard ) )
		return ALPHABETA_ILLEGAL;

    // Sort the moves according to History heuristic values
    if (sort) CutoffHist.SortMoveList( &movegen, theBoard->GetCurrentPlayer() );


    // OK, now, get ready to search
    Board newBoard;
    int bestSoFar;

	Move* pmov = NULL;

    // Case #1: We are searching a Max Node
    if ( nodeType == MAXNODE ) {
		bestSoFar = ALPHABETA_MINVAL;
		int currentAlpha = alpha;

		// Loop on the successors
		while( ( pmov = movegen.Next() ) != NULL ) {
			// Compute a board position resulting from the current successor
			newBoard.Clone( theBoard );
			newBoard.ApplyMove( pmov );

			// And search it in turn
			int movScore = BasicAlphaBetaTT( !nodeType, &newBoard, depth - 1, currentAlpha, beta, sort );
			// Ignore illegal moves in the alphabeta evaluation
			if ( movScore == ALPHABETA_ILLEGAL ) 
					continue;

			currentAlpha = max( currentAlpha, movScore );

			// Is the current successor better than the previous best?
			if ( movScore > bestSoFar ) {
				Bestmov.Set( pmov );
				bestSoFar = movScore;
				// Can we cutoff now?
				if ( bestSoFar >= beta ) {
					// Store this best move in the TransTable
					TransTable.StoreBoard( theBoard, bestSoFar, Move::EVALTYPE_LOWERBOUND, depth, MoveCounter, &Bestmov, NumRegularNodes );

					// Add this move's efficiency in the HistoryTable
					CutoffHist.AddCount( theBoard->GetCurrentPlayer(), &Bestmov );
					NumRegularCutoffs++;

					return bestSoFar;
				}
			}
		}
/*
		// Test for checkmate or stalemate
		// Both cases occur if and only if there is no legal move for MAX, i.e.,
		// if "bestSoFar" is ALPHABETA_MINVAL.  There are two cases: we
		// have checkmate ( in which case the score is accurate ) or stalemate ( in
		// which case the position should be re-scored as a draw with value 0.
		if ( bestSoFar <= ALPHABETA_MINVAL ) {
			// Can MIN capture MAX's king?  First, ask the machine to generate
			// moves for MIN
			newBoard.Clone( theBoard );
			if( newBoard.GetCurrentPlayer() == FromWhosePerspective ) 
				newBoard.SwitchSides();

			// And if one of MIN's moves is a king capture, indicating that the
			// position is illegal, we have checkmate and must return MINVAL.  We
			// add the depth simply to "favor" delaying tactics: a mate in 5 will
			// score higher than a mate in 3, because the likelihood that the
			// opponent will miss it is higher; might as well make life difficult!
			if ( !movegen.ComputeLegalMoves( &newBoard ) )
				return bestSoFar + depth;
				else
				return 0;
		}
*/
    } else { // Case #2: Min Node
		bestSoFar = ALPHABETA_MAXVAL;
		int currentBeta = beta;
		while( ( pmov = movegen.Next() ) != NULL ) {
			newBoard.Clone( theBoard );
			newBoard.ApplyMove( pmov );

			int movScore = BasicAlphaBetaTT( !nodeType, &newBoard, depth - 1, alpha, currentBeta, sort );
			if ( movScore == ALPHABETA_ILLEGAL )
				continue;
			currentBeta = min( currentBeta, movScore );
			if ( movScore < bestSoFar ) {
				Bestmov.Set( pmov );
				bestSoFar = movScore;
				// Cutoff?
				if ( bestSoFar <= alpha ) {
					TransTable.StoreBoard( theBoard, bestSoFar, Move::EVALTYPE_UPPERBOUND, depth, MoveCounter, &Bestmov, NumRegularNodes);
					CutoffHist.AddCount( theBoard->GetCurrentPlayer(), pmov );
					NumRegularCutoffs++;
			
					return bestSoFar;
				}
			}
		}
/*
		// Test for checkmate or stalemate
		if ( bestSoFar >= ALPHABETA_MAXVAL ) {
			// Can MAX capture MIN's king?
			newBoard.Clone( theBoard );
			if( newBoard.GetCurrentPlayer() != FromWhosePerspective )
				newBoard.SwitchSides();
			if ( !movegen.ComputeLegalMoves( &newBoard ) )
				return bestSoFar + depth;
				else
				return 0;
		}
*/
    }

    // If we haven't returned yet, we have found an accurate minimax score
    // for a position which is neither a checkmate nor a stalemate
    TransTable.StoreBoard( theBoard, bestSoFar, Move::EVALTYPE_ACCURATE, depth, MoveCounter, &Bestmov, NumRegularNodes );
    return bestSoFar;
}


#define fup  mov.MoveEvaluation
#define flow mov.MovingPiece

int AISearchAgent::BasicAlphaBetaTTNew(bool nodeType, Board *theBoard, int depth, int alpha, int beta, bool sort)
{
	int g;
	Move* c;
	Move mov;
	MoveListGenerator movegen;
	Board newBoard;

	NumRegularNodes++;

    if ( TransTable.LookupBoard( theBoard, &mov ) && ( mov.SearchDepth >= depth ) ) {
		if ( flow >= beta ) {
			NumRegularTTHits++;
			return flow;
		}
		if ( fup <= alpha ) {
			NumRegularTTHits++;
			return fup;
		}
		alpha = max( alpha, flow );
		beta  = min( beta,  fup  );
    }


	if (depth == 0) return Evaluator->EvaluateComplete(theBoard, FromWhosePerspective);

	if ( !movegen.ComputeLegalMoves( theBoard ) )
		return AISearchAgent::ALPHABETA_ILLEGAL;
    // Sort the moves according to History heuristic values
    if (sort) CutoffHist.SortMoveList( &movegen, theBoard->GetCurrentPlayer() );

	if (nodeType == AISearchAgent::MAXNODE) {
		g = AISearchAgent::ALPHABETA_MINVAL;

		while ((g < beta) && (c = movegen.Next())) {
			newBoard.Clone( theBoard );
			newBoard.ApplyMove( c );
			int newscore = BasicAlphaBetaTTNew(!nodeType, &newBoard, depth-1, alpha, beta, sort);
			g = max(g, newscore);
			alpha = max(alpha, g);
			if (g >= beta) 
				CutoffHist.AddCount( theBoard->GetCurrentPlayer(), c );
		}
	} else { //minnode
		g = AISearchAgent::ALPHABETA_MAXVAL;

		while ((g > alpha) && (c = movegen.Next())) {
			newBoard.Clone( theBoard );
			newBoard.ApplyMove( c );
			int newscore = BasicAlphaBetaTTNew(!nodeType, &newBoard, depth-1, alpha, beta, sort);
			g = min(g, newscore);
			beta = min(beta, newscore);
			if (g <= alpha) 
				CutoffHist.AddCount( theBoard->GetCurrentPlayer(), c );
		}
	}

	if (g < beta)  fup  = g;
	if (g > alpha) flow = g;
	TransTable.StoreBoard( theBoard, fup, Move::EVALTYPE_ACCURATE, depth, 0, &mov, flow);

	return g;
}

#undef fup
#undef flow

int AISearchAgent::MT(bool nodeType, Board *theBoard, int depth, int esti)
{
	int nfup  = ALPHABETA_MAXVAL;
	int nflow = ALPHABETA_MINVAL;
	Move mov;
	int g;

	NumRegularNodes++;

	if (depth == 0) {
		if ( TransTable.LookupBoard( theBoard, &mov ) && ( mov.SearchDepth >= depth ) ) {
			if (mov.EVALTYPE_LOWERBOUND) {
				nflow = mov.MoveEvaluation;
				g = nflow;
			}
			if (mov.EVALTYPE_UPPERBOUND) {
				nfup  = mov.MoveEvaluation;
				g = nfup;
			}
		} else {
			g = Evaluator->EvaluateComplete(theBoard, FromWhosePerspective);
		}
	} else {
		MoveListGenerator movegen;
		if ( !movegen.ComputeLegalMoves( theBoard ) )
			return ALPHABETA_ILLEGAL;

		// Sort the moves according to History heuristic values
		CutoffHist.SortMoveList( &movegen, theBoard->GetCurrentPlayer() );


		// OK, now, get ready to search
		Board newBoard;

		Move* c = NULL;
		int cfup  = ALPHABETA_MAXVAL;
		int cflow = ALPHABETA_MINVAL;
		int newg;

		// Case #1: We are searching a Max Node
		if ( nodeType == MAXNODE ) {
			g = ALPHABETA_MINVAL;

			// Loop on the successors
			while( ( g < esti ) && (( c = movegen.Next() ) != NULL )) {
				// Compute a board position resulting from the current successor
				newBoard.Clone( theBoard );
				newBoard.ApplyMove( c );
				if ( TransTable.LookupBoard( &newBoard, &mov ) && ( mov.SearchDepth >= depth-1 ) ) {
					if (mov.EVALTYPE_UPPERBOUND) cfup = mov.MoveEvaluation;
				}
				if (cfup >= esti) {
					newg = MT(!nodeType, &newBoard, depth - 1, esti);
				} else newg = cfup;
				g = max(g, newg);
				if (g >= esti) {
					CutoffHist.AddCount( theBoard->GetCurrentPlayer(), c );
				}
			}
		} else { // MINNODE
			g = ALPHABETA_MAXVAL;

			// Loop on the successors
			while( ( g >= esti ) && (( c = movegen.Next() ) != NULL )) {
				// Compute a board position resulting from the current successor
				newBoard.Clone( theBoard );
				newBoard.ApplyMove( c );
				if ( TransTable.LookupBoard( &newBoard, &mov ) && ( mov.SearchDepth >= depth-1 ) ) {
					if (mov.EVALTYPE_LOWERBOUND) cflow = mov.MoveEvaluation;
				}
				if (cflow < esti) {
					newg = MT(!nodeType, &newBoard, depth - 1, esti);
				} else newg = cflow;
				g = min(g, newg);
				if (g < esti) {
					CutoffHist.AddCount( theBoard->GetCurrentPlayer(), c );
				}
			}
		}
	}

	if (g >= esti) {
		nflow = g; TransTable.StoreBoard(theBoard, nflow, Move::EVALTYPE_LOWERBOUND, depth, 0, &mov, 0);
	} else {
		nfup  = g; TransTable.StoreBoard(theBoard, nfup,  Move::EVALTYPE_UPPERBOUND, depth, 0, &mov, 0);
	}

	return g;
}




void AISearchAgent::TestAgent(Board* theBoard)
{

	FromWhosePerspective = theBoard->CurrentPlayer;

	int MaxDepth = 6;

	int i;
	int g;
	int upper;
	int lower;
	int eval = 0;
	int eval2 = 0;
	int gam;
	Move mymove;
	AISearchAgentNegaScout NS(NULL);  

	TransTable.Clear();
	CutoffHist.Forget(); 
//	for ( i = 2; i <= 2; i++) {
	for ( i = 1; i <= MaxDepth; i++) {
		if (i % 4 == 0) CutoffHist.Forget(); 
		cout << setw(2) << i << " NegaScout: " << setw(5) << NS.NegaScoutDriver(theBoard, 0, i, &mymove);
		cout << " (" << setw(7) << NS.NumRegularNodes << " - col: " << NS.NumRegularHashCollisions << " )" << endl;
		string mystring;
		Board::SquareToCoord(mymove.SourceSquare, mymove.DestinationSquare, &mystring);
		cout << mystring << endl;
		cout << BestMoves::Get(i, theBoard, &NS.TransTable) << endl;

	}

	cout << endl;

/*

	TransTable.Clear();
	CutoffHist.Forget(); 
	for ( i = 1; i <= MaxDepth; i++) {
		NumRegularNodes = 0;
		if (i % 4 == 0) CutoffHist.Forget(); 
		cout << setw(2) << i << " BasicAlphaBetaTT:    " << setw(5) << BasicAlphaBetaTT(MAXNODE, theBoard, i, ALPHABETA_MINVAL, ALPHABETA_MAXVAL, true);
		cout << " (" << setw(7) << NumRegularNodes << ")" << endl;
	}

	cout << endl;

	TransTable.Clear();
	CutoffHist.Forget(); 
	for ( i = 1; i <= MaxDepth; i++) {
		NumRegularNodes = 0;
		if (i % 4 == 0) CutoffHist.Forget(); 
		cout << setw(2) << i << " BasicAlphaBetaTTNew: " << setw(5) << setw(5) << BasicAlphaBetaTTNew(MAXNODE, theBoard, i, ALPHABETA_MINVAL, ALPHABETA_MAXVAL, true);
		cout << " (" << setw(7) << NumRegularNodes << ")" << endl;
	}

	cout << endl;
*/
/*
	TransTable.Clear();
	CutoffHist.Forget(); 
	for ( i = 1; i <= MaxDepth; i++) {
		NumRegularNodes = 0;
		if (i % 4 == 0) CutoffHist.Forget(); 
		g = ALPHABETA_MAXVAL;
		int myesti;
		do {
			myesti = g;
			g = MT(MAXNODE, theBoard, i, myesti);
		}
		while (g != myesti);

		cout << setw(2) << i << " MT-SSS: " << setw(5) << g;
		cout << " (" << setw(7) << NumRegularNodes << ")" << endl;
	}

	cout << endl;

	TransTable.Clear();
	CutoffHist.Forget(); 
	for ( i = 1; i <= MaxDepth; i++) {
		NumRegularNodes = 0;
		if (i % 4 == 0) CutoffHist.Forget(); 
		g = ALPHABETA_MINVAL;
		int myesti;
		do {
			myesti = g;
			g = MT(MAXNODE, theBoard, i, myesti + 1);
		}
		while (g != myesti);

		cout << setw(2) << i << " MT-DUAL: " << setw(5) << g;
		cout << " (" << setw(7) << NumRegularNodes << ")" << endl;
	}

	cout << endl;

*/
    TransTable.Clear();
	CutoffHist.Forget(); 
	for ( i = 1; i <= MaxDepth; i++) {
		NumRegularNodes = 0;
		if (i % 4 == 0) CutoffHist.Forget(); 
		g = eval;
		upper  = ALPHABETA_MAXVAL;
		lower = ALPHABETA_MINVAL;
		do {
			if (g == ALPHABETA_MINVAL) gam = g + 1; else gam = g;
			g = BasicAlphaBetaTT(MAXNODE, theBoard, i, gam - 1, gam, true);
			if ( g < gam ) upper = g; else lower = g;
		}
		while (upper < lower);

		cout << setw(2) << i << " MTBasicAlphaBetaTT: " << setw(5) << setw(5) << g;
		cout << " (" << setw(7) << NumRegularNodes << ")" << endl;
	}

	cout << endl;

	TransTable.Clear();
	CutoffHist.Forget(); 
	for ( i = 1; i <= MaxDepth; i++) {
		NumRegularNodes = 0;
		if (i % 4 == 0) CutoffHist.Forget(); 

		int beta;
		Move Mov;
		int currentEstimate = eval;
		int upperbound = ALPHABETA_MAXVAL;
		int lowerbound = ALPHABETA_MINVAL;

		// This is the trick: make repeated calls to alphabeta, zeroing in on the
		// actual minimax value of theBoard by narrowing the bounds
		do {
			if ( currentEstimate == lowerbound ) {
				beta = currentEstimate + 1;
			} else
				beta = currentEstimate;

			currentEstimate = BasicAlphaBetaTTNew(MAXNODE, theBoard, i, beta - 1, beta, true );
			
			if ( currentEstimate < beta )
				upperbound = currentEstimate;
				else
				lowerbound = currentEstimate;

		} while ( lowerbound < upperbound );		

		cout << setw(2) << i << " MTBasicAlphaBetaTTNew: " << setw(5)  << currentEstimate << " " << setw(5) << lowerbound << " " << upperbound;
		cout << " (" << setw(7) << NumRegularNodes << ")" << endl;
	}

	cout << endl;

	TransTable.Clear();
	CutoffHist.Forget(); 
	for ( i = 1; i <= MaxDepth; i++) {
		NumRegularNodes = 0;
		if (i % 4 == 0) CutoffHist.Forget(); 
		cout << setw(2) << i << " BasicAlphaBetaTTNew: " << setw(5) << setw(5) << BasicAlphaBetaTTNew(MAXNODE, theBoard, i, ALPHABETA_MINVAL, ALPHABETA_MAXVAL, true);
		cout << " (" << setw(7) << NumRegularNodes << ")" << endl;
	}

	cout << endl;
		
	CutoffHist.Forget(); 
	for ( i = 1; i <= MaxDepth; i++) {
		NumRegularNodes = 0;
		if (i % 4 == 0) CutoffHist.Forget(); 
		cout << setw(2) << i << " BasicAlphaBeta:      " << setw(5) << BasicAlphaBeta(MAXNODE, theBoard, i, ALPHABETA_MINVAL, ALPHABETA_MAXVAL, true);
		cout << " (" << setw(7) << NumRegularNodes << ")" << endl;
	}
	
	cout << endl << "UNSORTED" << endl;

	CutoffHist.Forget(); 
	for ( i = 1; i <= MaxDepth; i++) {
		NumRegularNodes = 0;
		if (i % 4 == 0) CutoffHist.Forget(); 
		cout << setw(2) << i << " MiniMax:             " << setw(5) << MiniMax(MAXNODE, theBoard, i);
		cout << " (" << setw(7) << NumRegularNodes << ")" << endl;
	}

	TransTable.Clear();
	CutoffHist.Forget(); 
	for ( i = 1; i <= MaxDepth; i++) {
		NumRegularNodes = 0;
		if (i % 4 == 0) CutoffHist.Forget(); 
		cout << setw(2) << i << " BasicAlphaBetaTT:    " << setw(5) << BasicAlphaBetaTT(MAXNODE, theBoard, i, ALPHABETA_MINVAL, ALPHABETA_MAXVAL);
		cout << " (" << setw(7) << NumRegularNodes << ")" << endl;
	}

	cout << endl;

	TransTable.Clear();
	CutoffHist.Forget(); 
	for ( i = 1; i <= MaxDepth; i++) {
		NumRegularNodes = 0;
		if (i % 4 == 0) CutoffHist.Forget(); 
		cout << setw(2) << i << " BasicAlphaBetaTTNew: " << setw(5) << setw(5) << BasicAlphaBetaTTNew(MAXNODE, theBoard, i, ALPHABETA_MINVAL, ALPHABETA_MAXVAL);
		cout << " (" << setw(7) << NumRegularNodes << ")" << endl;
	}

	cout << endl;
		
	CutoffHist.Forget(); 
	for ( i = 1; i <= MaxDepth; i++) {
		NumRegularNodes = 0;
		if (i % 4 == 0) CutoffHist.Forget(); 
		cout << setw(2) << i << " BasicAlphaBeta:      " << setw(5) << BasicAlphaBeta(MAXNODE, theBoard, i, ALPHABETA_MINVAL, ALPHABETA_MAXVAL);
		cout << " (" << setw(7) << NumRegularNodes << ")" << endl;
	}

		

}

void AISearchAgent::ShowStats() 
{
}
