//
// Move.h - An encapsulation of a chess move and its consequences
//
// Purpose: This class is used all over the place.  It contains a move's
// source and target squares, a type identifier ( i.e., a normal move, a pawn
// promotion, etc. ) and a score, whether an actual evaluation of the position
// which would result from the move or a value taken from the history table.
//
// History
// 11.06.00 Creation ( Franois Dominic Larame )
// 09.07.00 Added fields MovingPiece and CapturedPiece; while not absolutely
//          needed, they do accelerate move processing and help to make code
//          easier to understand, so I gladly keep them around as optimizations
//          ( Franois Dominic Larame )
// 14.08.00 Added "search depth" field, so that we can determine whether a
//          transposition table entry should be used or not. ( Franois Dominic 
//          Larame )
// 13.02.03 Converted to C++ ( Clemens Kutz )
//
//////////////////////////////////////////////////////////////////////

#if !defined( AFX_MOVE_H__9F7C7383_EB27_42BD_9294_680CF5015802__INCLUDED_ )
#define AFX_MOVE_H__9F7C7383_EB27_42BD_9294_680CF5015802__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

class Move  
{
/***************************************************************************
 * Static Constants
 **************************************************************************/
public:
 /**
   * einfacher Zug
   */
  static const int MOVE_NORMAL;
  /**
   * A schlgt Figur von B
   */
  static const int MOVE_CAPTURE_ORDINARY;
  /**
   * A schlgt mit Bauern Figur von B ( en passant )
   */
  static const int MOVE_CAPTURE_EN_PASSANT;
  /**
   * kurze Rochade
   */
  static const int MOVE_CASTLING_KINGSIDE;
  /** lange Rochade
   */
  static const int MOVE_CASTLING_QUEENSIDE;
  
  static const int MOVE_RESIGN;
  /**
   * Patt
   */
  static const int MOVE_STALEMATE;
  /**
   * Schachmatt
	 * Der letzte Zug des Gewinners ist von diesem Typ.
   */
  static const int MOVE_CHECKMATE;
  /**
   * Bauern zum Pferd aufwerten
   */
  static const int MOVE_PROMOTION_KNIGHT;
  /**
   * Bauern zum Lufer aufwerten
   */
  static const int MOVE_PROMOTION_BISHOP;
  /**
   * Bauern zum Turm aufwerten
   */  
  static const int MOVE_PROMOTION_ROOK;
  /**
   * Bauern zur Dame aufwerten
   */  
  static const int MOVE_PROMOTION_QUEEN;

  static const int MOVE_RATING_UNKNOWN;
  static const int MOVE_RATING_GOOD;
  static const int MOVE_RATING_POOR;

  /**
   * A pair of masks used to split the promotion and the non-promotion part of
   * a move type ID
   */
  static const int PROMOTION_MASK;
  static const int NO_PROMOTION_MASK;

  /**
   * Alphabeta may return an actual move potency evaluation, or an upper or
   * lower bound only ( in case a cutoff happens ).  We need to store this
   * information in the transposition table to make sure that a given
   * value is actually useful in given circumstances.
   */
  static const int EVALTYPE_ACCURATE;
  static const int EVALTYPE_UPPERBOUND;
  static const int EVALTYPE_LOWERBOUND;

  /**
   *  A sentinel value used to identify Move fields without valid data
   */
  static const int NULL_MOVE;

/***************************************************************************
 * Data Members
 **************************************************************************/
 public:
	/**
	  *  The moving piece; one of the constants defined by Board
	  */
	int MovingPiece;

	/**
      * The piece being captured by this move, if any; another Board constant
      */ 
	int CapturedPiece;

	/**
	  *  The squares involved in the move
	  */
	int SourceSquare, DestinationSquare;

	/**
	  * A type ID: is this a regular move, a capture, a capture AND promotion from
	  * Pawn to Rook, etc.  Move generation determines this, by definition; storing
	  * it here avoids having to "re-discover" the information in Board.ApplyMove
	  * at the cost of a few bytes
	  */
	int MoveType;

	/**
	  * An evaluation of the move's potency, either as a result of an alphabeta
	  * search of some kind or of a retrieval in the transposition table
	  */
	int MoveEvaluation;
	int MoveEvaluationType;
	int SearchDepth;
  
/***************************************************************************
 * Constructor / Destructor
 **************************************************************************/
public:
	/**
	  * only calls Reset()
	  */
	Move();
	virtual ~Move();

	bool operator<( const Move& a ) const;


/***************************************************************************
 * Operations
 **************************************************************************/
public:
	istream& read( istream& is );
	ostream& write( ostream& os ) const;
	/**
	  * Kopiert jedes Attribut des Objektes target
	  * in dieses Objekt
	  */  
	bool Set( Move *target );
	/**
	  * Check whether two Move objects contain the same data ( not necessarily
	  * whether they are the same object in memory )
	  */
	bool Equals( Move *target );
	/**
	  * Initializes Data Members with NULL_MOVE
	  */
	bool Reset();
	/** 
	  * Display the move on standard output<br />
	  * Move: [ SOURCESQUARE, DESTSQUARE ] MOVETYPE
	  */ 
	void Print();
	void ThinkPrint();
	/**
	  * Display the Move in Standard Algebraic Notation
		*/
	bool PrintToSAN( string *s );
	/**
	  * Check the member variables integrity
		*/
	bool CheckIntegrity();
};

#endif // !defined( AFX_MOVE_H__9F7C7383_EB27_42BD_9294_680CF5015802__INCLUDED_ )
